# -*- coding: utf-8 -*-

"""
certifi.py
~~~~~~~~~~

This module returns the installation location of cacert.pem or its contents.
"""
import os
import sys

cacert_path = os.path.join(os.path.dirname(__file__), "cacert.pem")

if sys.platform == 'linux':

    import distro
    distrib = distro.linux_distribution(False)[0]
    if distrib in ('debian', 'ubuntu', 'linuxmint', 'raspbian'):
        cacert_path = '/etc/ssl/certs/ca-certificates.crt'
    elif distrib in ('rhel', 'centos', 'fedora'):
        cacert_path = '/etc/pki/tls/certs/ca-bundle.crt'
else:
    def read_certs(cacert_path):
        with open(cacert_path, 'r') as file_cacert:
            return file_cacert.read()

    def write_certs(cacert_path):
        with open(cacert_path, 'w') as file_cacert:
            for cert in get_certs_from_certstore():
                file_cacert.write(cert)

    if sys.platform == "win32":

        def get_certs_from_certstore():
            import ssl
            ssl_context = ssl.create_default_context()
            ssl_context.load_default_certs()
            return [ssl.DER_cert_to_PEM_cert(der_cert) for der_cert in ssl_context.get_ca_certs(binary_form=True)]

        try:
            win_cacert_path_dir = os.path.join(os.environ['LOCALAPPDATA'], '.certifi')
            win_cacert_path = os.path.join(win_cacert_path_dir, 'cacert.pem')
            if not os.path.exists(win_cacert_path) or read_certs(win_cacert_path) != ''.join(get_certs_from_certstore()):
                os.makedirs(win_cacert_path_dir, exist_ok=True)
                write_certs(win_cacert_path)
            cacert_path = win_cacert_path
        except:
            pass
    elif sys.platform == 'darwin':

        from Security import (
            errSecSuccess, kSecClass, kSecReturnRef, kSecMatchLimit, kSecMatchLimitAll,
            SecItemCopyMatching, kSecClassCertificate, kSecMatchTrustedOnly,
            SecItemExport, kSecFormatUnknown,
            SecTrustCopyAnchorCertificates)

        def get_certs_from_certstore():
            certs = get_trusted_certs()
            certs.extend(get_system_roots())
            return_code, pem_data = SecItemExport(certs, kSecFormatUnknown, 0, None, None)
            return [cert.lstrip("\n")+'-----END CERTIFICATE-----\n' for cert in bytes(pem_data).decode().split('-----END CERTIFICATE-----') if cert.lstrip("\n")] if return_code == errSecSuccess else []

        def get_trusted_certs():
            query = {
                kSecClass: kSecClassCertificate,
                kSecReturnRef: True,
                kSecMatchLimit: kSecMatchLimitAll,
                kSecMatchTrustedOnly: True
            }
            result_code, result = SecItemCopyMatching(query, None)
            return list(result) if result_code == errSecSuccess else []

        def get_system_roots():
            result_code, result = SecTrustCopyAnchorCertificates(None)
            return list(result) if result_code == errSecSuccess else []

        try:
            macos_cacert_path_dir = os.path.join(os.path.expanduser("~"), ".config", ".certifi")
            macos_cacert_path = os.path.join(macos_cacert_path_dir, 'cacert.pem')
            if not os.path.exists(macos_cacert_path) or read_certs(macos_cacert_path) != ''.join(get_certs_from_certstore()):
                os.makedirs(macos_cacert_path_dir, exist_ok=True)
                write_certs(macos_cacert_path)
            cacert_path = macos_cacert_path
        except:
            pass

def where():
    return cacert_path

def contents():
    with open(where(), 'r', encoding='ascii') as data:
        return data.read()


if __name__ == '__main__':
    print(contents())
