#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2023  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##
from flask import Flask

from waptserver.config import load_config
import logging
import requests
import ujson
import time

logger = logging.getLogger('waptserver')

# allow chunked uploads when no nginx reverse proxy server server (see https://github.com/pallets/flask/issues/367)


class FlaskApp(Flask):
    def __init__(self, *args, **kwargs):
        super(FlaskApp, self).__init__(*args, **kwargs)
        self.conf = load_config()

    def request_context(self, environ):
        # it's the only way I've found to handle chunked encoding request (otherwise flask.request.stream is empty)
        environ['wsgi.input_terminated'] = 1
        return super(FlaskApp, self).request_context(environ)


app = FlaskApp(__name__, static_folder='./templates/static')

### Websockets SID in memory
# a dictionary with host uuid as keys and (connection_sid, certificate)
app.ws_connections = {}

# constant for item position in  app.ws_connections list items
WS_SID = 0
WS_TIMESTAMP = 1

# if server owns the socketio SID dict, return directly from it
# else query using http the local socketio server
# cache returned dict in app for 10 seconds.
def get_ws_connections(uuids=None):
    ws_connections = app.ws_connections
    if ws_connections is None:
        if hasattr(app,'ws_connections_cache') and time.time() - app.ws_connections_cache_timestamp < 10:
            ws_connections = app.ws_connections_cache
        else:
            # TODO: caching ... ?

            url_root = 'http://127.0.0.1:%s' % (app.conf['waptserver_port'],)

            if uuids is None:
                r = requests.get('%s/api/v3/hosts_sid' % (url_root,))
            else:
                r = requests.get('%s/api/v3/hosts_sid?host_uuids=%s' % (url_root,','.join(uuids)))
            if r.status_code  == 200:
                js = ujson.loads(r.content)
                ws_connections = js.get('result')
                # cache if get all
                if uuids is None:
                    app.ws_connections_cache = ws_connections
                    app.ws_connections_cache_timestamp = time.time()
            else:
                return {}

    if uuids is None:
        return ws_connections
    else:
        result = {}
        for uuid in uuids:
            ws = ws_connections.get(uuid)
            if ws and ws[WS_SID]:
                result[uuid] = ws
        return result

def get_ws_connection(uuid):
    ws_connection = get_ws_connections().get(uuid)
    return ws_connection

def host_is_reachable(uuid):
    ws = get_ws_connection(uuid)
    if not ws:
        return False
    return ws[WS_SID] is not None

def connected_uuids(from_uuids = None):
    return [uuid for uuid,ws in get_ws_connections().items() if ws[WS_SID] is not None and (from_uuids is None or uuid in from_uuids) ]

def append_ws_status(host_dict,uuid_keyname = 'uuid'):
    ws = get_ws_connection(host_dict[uuid_keyname])
    if ws:
        host_dict['listening_address'] = ws[WS_SID]
        host_dict['listening_timestamp'] = ws[WS_TIMESTAMP]
        host_dict['listening_protocol'] = 'websockets'
    else:
        host_dict['listening_address'] = None
        host_dict['listening_timestamp'] = None
        host_dict['listening_protocol'] = None
    host_dict['reachable'] = 'OK' if host_dict['listening_address'] else 'DISCONNECTED'

