#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2023  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import sys
import os
import logging
import traceback
import json
import ipaddress
import datetime
import glob

import hashlib
import threading

from flask import Response
from itsdangerous import URLSafeTimedSerializer

from waptserver.app import app

from waptutils import jsondump
from waptcrypto import SSLPrivateKey,SSLCertificate,SSLCRL

from waptpackage import WaptLocalRepo

from licencing import WaptActivatedLicence

oldsysargv = sys.argv
try:
    import waptlicences
except ImportError:
    waptlicences = None

sys.argv = oldsysargv

from operator import itemgetter
from waptutils import __version__

logger = logging.getLogger()

__all__ = [
    'make_response',
    'make_response_from_exception',
    'get_server_uuid',
    'get_secured_token_generator',
    'get_revision_hash',
    'get_wapt_full_version',
]

try:
    wapt_root_dir = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            '..'))
except:
    wapt_root_dir = 'c:/tranquilit/wapt'

def get_revision_hash():
    fn = os.path.join(wapt_root_dir, 'revision.txt')
    if os.path.isfile(fn):
        with open(fn, 'r') as f:
            return f.read()
    else:
        fn = os.path.join(wapt_root_dir, 'waptserver', 'VERSION')
        if os.path.isfile(fn):
            with open(fn, 'r') as f:
                return f.read()
        else:
            return ''

def get_wapt_full_version():
    fn = os.path.join(wapt_root_dir, 'version-full')
    if os.path.isfile(fn):
        with open(fn, 'r') as f:
            return f.read()
    else:
        return __version__

# API V2 #####


def make_response(result={}, success=True, error_code='', msg='', status=200, request_time=None, headers = None):
    data = dict(
        success=success,
    )

    if msg:
        data['msg']=msg
    if not success:
        data['error_code'] = error_code
    else:
        data['result'] = result
    if request_time:
        data['request_time'] = request_time
    return Response(
        response=jsondump(data),
        status=status,
        mimetype='application/json',
        headers = headers)


def make_response_from_exception(exception, error_code='', status=200):
    """Return a error flask http response from an exception object
        success : False
        msg : message from exception
        error_code : classname of exception if not provided
        status: 200 if not provided
    """
    if not error_code:
        error_code = type(exception).__name__.lower()
    data = dict(
        success=False,
        error_code=error_code
    )

    logger.info(traceback.format_exc())

    data['msg'] = 'Error on server:\n%s' % (repr(exception))
    return Response(
        response=jsondump(data),
        status=status,
        mimetype='application/json')


def get_server_uuid():
    """Returns this server UUID as configured in configuration file waptserver.ini
    """
    return app.conf.get('server_uuid', None)



def get_secondary_repo(ip, default_url='/', remove_wapt=True,repo_name=None):
    """Returns the secondary repo URL given an agent's source ip
    Rules are taken from json file in root wapt repor folder/rules.json
    Only 'AGENT IP' rules are considered.
    if default_url is a full http url, the protocol of result (http or hhtps) is
        copied from default_url

    Args:
        ip (str):           source ip of agent
        default_url:        URL to return if no match is found in rules.
        remove_wapt (bool): if True, remove the /wapt of the returned url.
        repo_name:          if not empty, only rules for this kind of repo are considered.

    Returns
        str:  URL of secondary repo matching the rules or default_url

    """

    rules_json = os.path.abspath(os.path.join(app.conf['wapt_folder'], os.pardir, 'rules.json'))
    if os.path.isfile(rules_json):
        with open(rules_json, 'r') as rules_file:
            try:
                rules = json.loads(rules_file.read())
            except:
                rules = {}
    else:
        return default_url

    for item in sorted(rules, key=itemgetter('sequence')):
        if not item['condition'] == "AGENT IP":
            continue

        if repo_name :
            if not repo_name in item['repositories']:
                continue
        try:
            ip_network = ipaddress.ip_network(item['value'])
        except ValueError:
            continue

        if ip :
            member = bool(ipaddress.ip_address(ip) in ip_network)
        else:
            member = False

        #fallback non testable

        if item['negation']:
            member = not(member)

        if member:
            res = item['repo_url'] if not remove_wapt or not item['repo_url'].endswith('/wapt') else item['repo_url'][:-4]
            if not res.endswith('/'):
                res = res + '/'

            #follow protocol of the server connection
            if '://' in default_url:
                proto = default_url.split('://',1)[0]
                res_without_proto = res.split('://',1)[1]
                res = proto + '://' + res_without_proto

            return res
    return default_url if not remove_wapt or not default_url.endswith('/wapt') else default_url[:-4]


def get_secured_token_generator(secret_key=None):
    """Get a timed encrypted token generator.

    Args:
        secret_key (str or None):
                            secret symetric encryption key for token.
                            if None, use app.conf.get('token_secret_key')
    Returns:
        URLSafeTimedSerializer
    """

    if not secret_key:
        secret_key = app.conf.get('token_secret_key')
    return URLSafeTimedSerializer(secret_key)


class EWaptLicenceDirectorySSLCertificateNotExist(Exception):
    pass

class LicenceStatus:
    lsUnknown=0
    lsValid=1
    lsNotJson=2
    lsBadServer=3
    lsBadSignature=4
    lsNotYetValid=5
    lsExpired=6
    lsDuplicated=7

    text = {
        0:'Unknown',
        1:'Valid',
        2:'Not Json',
        3:'Bad Server',
        4:'Bad Signature',
        5:'Not Yet Valid',
        6:'Expired',
        7:'Duplicated'
    }

class WaptLicencesList():

    def __init__(self, conf=None):
        self._conf = conf
        self._json_path = None
        self._ssl_certificate = None

        self.hosts_count = 0
        self.licences_list = None
        self.timestamp_json = None
        self.server_uuid = None

    @property
    def conf(self):
        return self._conf

    @conf.setter
    def conf(self,value):
        self._conf = value
        self._json_path = None
        self._ssl_certificate = None

        self.hosts_count = 0
        self.licences_list = None
        self.timestamp_json = None
        self.server_uuid = None

    @property
    def json_path(self):
        if not self._json_path:
            self._json_path = os.path.join(os.path.dirname(self.conf['wapt_folder']), 'licences.json')
        return self._json_path

    def is_enterprise(self):
        self.load_licences()
        return waptlicences and waptlicences.is_enterprise()

    def check_licence_count(self,licences_data):
        if not waptlicences:
            return 0
        aserver_uuid = self.conf.get('server_uuid', '') or '-'
        data_check = waptlicences.check_valid_licences_count(licences_data or b"[]",aserver_uuid,0)
        self.server_uuid = aserver_uuid
        self.hosts_count = data_check['count']
        self.licenced_to = data_check['licenced_to']
        self.licences_list = data_check['licences']
        return self.hosts_count

    def check_licence(self, licence_data):
        """Check that a licence is properly signed.

        If the licence has a server_uuid, certificate used to check is
        the relicencing cert
        Else it is the TIS licencing cert

        Args:
            licence_data (str): json of signed licence object

        Returns: int
            0:'Valid',
            1:'Unknown',
            2:'Not Json',
            3:'Bad Server',
            4:'Bad Signature',
            5:'Not Yet Valid',
            6:'Expired',
            7:'Duplicated'
        """
        if not waptlicences:
            return 1
        if not isinstance(licence_data,str):
            licence_data = jsondump(licence_data)
        return waptlicences.check_licence_signature(licence_data,get_server_uuid())

    def check_time_validity_licence(self, licence_data):
        """Check that a licence is not expired.
        """
        return WaptActivatedLicence(licence_data_dict=licence_data).is_valid_until_from()

    def add_licence(self, licence = None):
        self.load_licences()
        if isinstance(licence, (str, bytes)):
            new_licence = json.loads(licence)
        else:
            new_licence = licence
        if new_licence['licence_nr'] in [l['licence_nr'] for l in self.licences_list]:
            return (None,LicenceStatus.lsDuplicated,'Licence %s already exists on server' % new_licence['licence_nr'])
        status = self.check_licence(new_licence)
        if status == LicenceStatus.lsValid:
            self.licences_list.append(new_licence)
            self.write_licences()
            return (new_licence, status, 'Activation of %(licenced_to)s-%(licence_nr)s OK' % new_licence)
        else:
            return (None, status, LicenceStatus.text[status])

    def load_licences(self):
        if self.timestamp_json is None or not os.path.isfile(self.json_path)\
                or self.timestamp_json < os.path.getmtime(self.json_path)\
                or self.licences_list is None or self.server_uuid != self.conf.get('server_uuid', ''):
            try:
                with open(self.json_path, 'r') as json_licences:
                    licences_data = json_licences.read()
            except (FileNotFoundError, json.decoder.JSONDecodeError):
                self.licences_list = []
                self.write_licences()
                licences_data = '[]'
            self.check_licence_count(licences_data)
            self.timestamp_json = os.path.getmtime(self.json_path)

    def write_licences(self):
        with open(self.json_path, 'w') as json_licences:
            json_licences.write(json.dumps(self.licences_list))
        # be sure to refresh valid licences
        self.load_licences()

    def remove_licences(self, licence_nr_list):
        self.load_licences()
        licence_to_remove = []
        for licence_nr in licence_nr_list:
            licence = self.get_licence(licence_nr)
            if licence is not None:
                licence_to_remove.append(licence)
                licence = None

        self.licences_list=[licence for licence in self.licences_list if licence not in licence_to_remove]
        self.write_licences()
        return licence_to_remove

    def remove_licence(self, licence_nr = None):
        if licence_nr is None:
            return None
        else:
            return self.remove_licences([licence_nr])[0]

    def get_licences(self):
        self.load_licences()
        return self.licences_list

    def get_licence(self, licence_nr = None):
        self.load_licences()
        for licence in self.licences_list:
            if licence['licence_nr'] == licence_nr:
                return licence
        return None

licences_list = WaptLicencesList()


class PostDataStream(object):
    """A n object to read post data from request in chunks
    """
    class DoHash(threading.Thread):
        def __init__(self):
            self.hash = hashlib.sha256()

        def update(self,chunk: bytes):
            self.hash.update(chunk)


    def __init__(self, stream, chunk_size=10*1024*1024, calc_sha256=False):
        self.stream = stream
        self.chunk_size = chunk_size
        self.hash = hashlib.sha256()

    def save(self, target):
        with open(target, "wb") as f:
            chunk = self.stream.read(self.chunk_size)
            while len(chunk) > 0:
                self.hash.update(chunk)
                f.write(chunk)
                chunk = self.stream.read(self.chunk_size)

    def sha256digest(self):
        return self.hash.hexdigest()

def update_all_crls(conf, force=True, crl_expire_days=10, crl_urls=None):
    """Sign CRL of hosts pki if needed (expire in less than 2 days, or force)
    Merge CRL of hosts, CRLs of all certificates in packages, and not expired CRLs in 'ssl_additional_crls' directory.
    Put the resulting CRLs bundle in 'ssl_client_crls' file for use by NGINX.

    Args:
        force (bool): TRue if resign hosts pki inconditionaly
        crl_expire_days (int): if expiration of crl is in less than crl_expire_days, sign the CRL.

    Returns:
        bool: True if crl has been updated, which means we must reload nginx.

    """
    try:
        all_crls = []
        crl = None
        crls_updated = False

        # sign crl of hosts.
        if (conf['clients_signing_key'] and conf['clients_signing_certificate'] and conf['clients_signing_crl'] and
                os.path.isfile(conf['clients_signing_key']) and os.path.isfile(conf['clients_signing_certificate'])):
            signing_key = SSLPrivateKey(conf['clients_signing_key'])
            signing_cert = SSLCertificate(conf['clients_signing_certificate'])
            logger.info('Checking CRL %s' % conf['clients_signing_crl'])
            crl = SSLCRL(conf['clients_signing_crl'], cakey=signing_key, cacert=signing_cert)
            if crl_expire_days > conf['clients_signing_crl_days']:
                crl_expire_days = conf['clients_signing_crl_days']
            if force or crl.crl is None or crl.next_update <= datetime.datetime.utcnow() - datetime.timedelta(days=crl_expire_days):
                crl.revoke_cert(crl_ttl_days=conf['clients_signing_crl_days'])
                crls_updated = True
                logger.info('Signing CRL %s, next update: %s' % (conf['clients_signing_crl'], crl.next_update))
            else:
                logger.info('Skip signing CRL %s, next update: %s' % (conf['clients_signing_crl'], crl.next_update))
            crl.save_as_pem(conf['clients_signing_crl']+'.new')
            # try to load it
            try:
                crl2 = SSLCRL(conf['clients_signing_crl']+'.new', cakey=signing_key, cacert=signing_cert)
                cn = crl2.authority_key_identifier
                os.unlink(conf['clients_signing_crl'])
                os.rename(conf['clients_signing_crl']+'.new', conf['clients_signing_crl'])
                all_crls.append(crl.issuer_dn.encode('utf8'))
                all_crls.append(crl.as_pem().strip())
                if not crl_urls is None:
                    crl_urls.append(crl.issuer_dn)
            except:
                logger.critical('unble to regenerate CRL %s. New CRL file corrupted' % conf['clients_signing_crl'])


        # merge all CRLs (signers CA + hosts CA) into a file for nginx
        if conf.get('ssl_client_crls') and conf.get('ssl_client_crls') != conf['clients_signing_crl']:
            logger.info('merging into %s' % conf.get('ssl_client_crls'))
            local_repo = WaptLocalRepo(conf['wapt_folder'])
            signers_ca = local_repo.get_certificates()
            signers_crls = signers_ca.crls
            for crl in signers_crls:
                crl_pem = crl.as_pem().strip()
                if not crl_pem in all_crls:
                    all_crls.append(crl.issuer_dn.encode('utf8'))
                    all_crls.append(crl_pem)
                    if not crl_urls is None:
                        crl_urls.append(crl.issuer_dn)

            # append manual CRL
            if conf.get('ssl_additional_crls') and os.path.isdir(conf.get('ssl_additional_crls')):
                logger.info('merging CRL from directory %s' % conf.get('ssl_additional_crls'))
                crls_fn = glob.glob(os.path.join(conf.get('ssl_additional_crls'),'*.crl'))
                for crl_fn in crls_fn:
                    logger.info('processing %s' % crl_fn)
                    try:
                        crl = SSLCRL(crl_fn)
                        logger.info('Issued by: %s, next update: %s' % ( crl.issuer_dn, crl.next_update ))
                        if crl.next_update > datetime.datetime.utcnow():
                            crl_pem = crl.as_pem().strip()
                            if not crl_pem in all_crls:
                                all_crls.append(crl.issuer_dn.encode('utf8'))
                                all_crls.append(crl_pem)
                                if not crl_urls is None:
                                    crl_urls.append(crl.issuer_dn)
                        else:
                            logger.info('Skipping %s' % crl_fn)
                    except Exception as e:
                        logger.critical('Error %s' % e)

            new_data = b'\n'.join(all_crls)
            # check if additional crls has changed
            if not crls_updated:
                with open(conf['ssl_client_crls'],'rb') as f:
                    old_data = f.read()
                    crls_updated = new_data != old_data

            if crls_updated:
                with open(conf['ssl_client_crls'],'wb') as f:
                    f.write(new_data)

        return crls_updated

    except Exception as e:
        logger.debug(traceback.format_exc())
        logger.critical('Unable to update CRLs %s: %s' % (conf['clients_signing_crl'], repr(e)))


