#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2023  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import functools
import logging
import datetime

from flask import request, session, after_this_request

from io import BytesIO as IO
import gzip

from waptserver.app import app
from waptserver.auth import check_auth, has_any_right, get_user_acls
from waptserver.model import wapt_db, LogsAPI
from waptserver.common import make_response
from waptserver.common import licences_list  # used as global directive below
import zlib

try:
    # i18n
    from flask_babel import Babel
    from flask_babel import gettext
except ImportError:
    gettext = (lambda s: s)
    Babel = None

_ = gettext

logger = logging.getLogger('waptserver')
authlogger = logging.getLogger('waptauth')


not_optimized_endpoints = [
    'waptagent_version',
    'host_tasks_status',
    'get_ad_groups',
    'get_ad_sites',
    'get_ad_ou_split',
    'host_data',
    'get_hosts',
    'audit_data',
    'wsus.windows_updates',
    'wsus.windows_products',
    'wsus.windows_updates_classifications',
    'packages_for_hosts',
    'enterprise.reporting_exec',
    'known_packages',
    'repositories.get_all_agentrepos',
    'repositories.get_sync_version',
    'repositories.get_all_rules',
    'get_all_users_acls',
    'known_signers_certificates',
    'enterprise.reporting_list',
    'usage_statistics',
    'repositories.get_createupdatefilesync',
    'repositories.get_sync_changelog',
    'licences'
    ]

no_logs_endpoints = [
    'get_auth_logs',
    'ping',
    ]

def authenticate(method='Basic'):
    """Sends a 401 response that enables basic auth"""
    return make_response(
        success=False,
        msg=_('You have to login with proper credentials'),
        status=401,
        headers = {'WWW-Authenticate': '%s realm="Login Required"' % method})
    #return Response(
    #    _('You have to login with proper credentials'), 401,
    #    {'WWW-Authenticate': '%s realm="Login Required"' % method})


def add_api_log(request=None, forbidden=False):
    if app.conf['authentication_logs'] and not request.endpoint in no_logs_endpoints and ((not app.conf['optimized_authentication_logs'] or not request.endpoint in not_optimized_endpoints)):
        user = session.get('user')
        data = None

        # decode json data
        if  (request.content_type == 'application/json' and
                not 'login' in request.endpoint and
                not 'change_password' in request.endpoint):
            raw_data = request.data

            if raw_data:
                try:
                    if request.headers.get('Content-Encoding') == 'gzip':
                        raw_data = zlib.decompress(raw_data)
                    data = raw_data
                except:
                    pass
            else:
                data = raw_data

        api_log = {
            "endpoint": request.endpoint,
            "args": request.args,
            "data": data,
            "user": user,
            "user_acls": session.get('user_acls'),
            "host": request.host,
            "remote_addr": request.headers.get('X-Real-Ip', request.headers.get('X-Forwarded-For')),
            "header_user_agent": request.headers.get('User-agent'),
            "method": request.method,
            "scheme": request.scheme,
            "path": request.path,
            "full_path": request.full_path,
            "created_on": datetime.datetime.utcnow().isoformat(),
            "updated_on": datetime.datetime.utcnow().isoformat(),
            "forbidden": forbidden,
            "authenticated": user is not None
        }
        must_close_on_exit = wapt_db.is_closed()
        if must_close_on_exit:
            wapt_db.connect()
        try:
            LogsAPI.insert(api_log).execute()
        finally:
            if must_close_on_exit:
                wapt_db.close()

def get_request_auth_http_status(request,session,methods=None,require_any_of=None):
    auth_result = check_auth(request=request, session=session, methods=methods)
    if not auth_result:
        add_api_log(request=request,forbidden=True)
        authlogger.info('Failed auth. Requires valid authentication for request %s from %s (%s)' % (request.path,request.headers.get('X-Forwarded-For'),request.user_agent))
        return 401

    user_acls = None
    if auth_result['auth_method'] == 'session':
        user_acls = session.get('user_acls', None)
    else:
        session.update(**auth_result)

    if user_acls is None:
        user_acls = get_user_acls(username=auth_result['user'])
        session['user_acls'] = user_acls

    if require_any_of and not has_any_right(user_acls, require_any_of):
        add_api_log(request=request, forbidden=True)
        authlogger.info('ERROR. User "%s" authenticated, but has not any of required %s acls for request %s from "%s" (%s)' % (auth_result['user'],require_any_of,request.path,request.headers.get('X-Forwarded-For'),request.user_agent))
        return 403

    authlogger.info('OK. User "%s" authenticated. %s for request %s from "%s" (%s)' % (auth_result['user'],auth_result,request.path,request.headers.get('X-Forwarded-For'),request.user_agent))
    return 200


def requires_auth(require_any_of=None, methods=None):
    """Flask route decorator which requires Basic Auth http header
    If not header, returns a 401 http status.

    Args:
        methods (list of str): list of authentication methods to check in turn
        require_any_of (None or list of str): requires that user has at least of these
                                                acls in any of its perimeters.

    Returns:
        decorator

    """
    def decorator(f):
        @functools.wraps(f)
        def _auth_wrapper(*args, **kwargs):
            http_status = get_request_auth_http_status(request,session,methods=methods,require_any_of=require_any_of)
            if http_status == 401:
                return authenticate()
            elif http_status == 403:
                return forbidden()
            result = f(*args, **kwargs)
            add_api_log(request=request)
            return result
        return _auth_wrapper
    return decorator


def allow_local(f):
    """Restrict access to localhost"""
    @functools.wraps(f)
    def decorated(*args, **kwargs):
        if request.remote_addr in ['127.0.0.1', '[::1]', '[::ffff:127.0.0.1]']:
            return f(*args, **kwargs)
        else:
            return forbidden()
    return decorated

def wads_enable(f):
    """Restrict wads if not enable"""
    @functools.wraps(f)
    def decorated(*args, **kwargs):
        if not app.conf['wads_enable']:
            return forbidden()
        else:
            return f(*args, **kwargs)

    return decorated

def reporting_enable(f):
    """Restrict reporting if not enable"""
    @functools.wraps(f)
    def decorated(*args, **kwargs):
        if not app.conf['reporting_enable']:
            return forbidden()
        else:
            return f(*args, **kwargs)

    return decorated

def add_logs(f):
    """Decorator for adding logs"""
    @functools.wraps(f)
    def decorated(*args, **kwargs):
        add_api_log(request=request)
        return f(*args, **kwargs)
    return decorated


def forbidden():
    """Sends a 403 response that enables basic auth"""
    return make_response(success=False,msg='forbidden',status=403)

def gzipped(f):
    @functools.wraps(f)
    def view_func(*args, **kwargs):
        @after_this_request
        def zipper(response):
            accept_encoding = request.headers.get('Accept-Encoding', '')

            if 'gzip' not in accept_encoding.lower():
                return response

            response.direct_passthrough = False

            if (response.status_code < 200 or
                response.status_code >= 300 or
                    'Content-Encoding' in response.headers):
                return response
            gzip_buffer = IO()
            gzip_file = gzip.GzipFile(mode='wb',
                                      fileobj=gzip_buffer)
            gzip_file.write(response.data)
            gzip_file.close()

            response.data = gzip_buffer.getvalue()
            response.headers['Content-Encoding'] = 'gzip'
            response.headers['Vary'] = 'Accept-Encoding'
            response.headers['Content-Length'] = len(response.data)

            return response

        return f(*args, **kwargs)

    return view_func


def wapt_db_readonly(f):
    @functools.wraps(f)
    def decorated(*args, **kwargs):
        if wapt_db and wapt_db.is_closed():
            wapt_db.connect()
        cnx = wapt_db.connection()
        b = cnx.readonly
        if not b:
            cnx.set_session(readonly=True)
        try:
            r = f(*args, **kwargs)
            return r
        finally:
            if not b:
                cnx.set_session(readonly=False)
            if wapt_db and not wapt_db.is_closed():
                wapt_db.close()
    return decorated


def require_wapt_db(f):
    @functools.wraps(f)
    def decorated(*args, **kwargs):
        if wapt_db and wapt_db.is_closed():
            logger.info('Connecting to waptdb before request. Count: %s' % len(wapt_db._in_use))
            wapt_db.connect()
        try:
            r = f(*args, **kwargs)
            return r
        finally:
            if wapt_db and not wapt_db.is_closed():
                wapt_db.close()
    return decorated


def forbidden_no_valid_licence():
    """Sends a 403 response that tells there's no valid enterprise licence"""
    return make_response(success=False,msg='Forbidden due to no valid enterprise licence.',status=403)


def requires_enterprise(f):
    """Flask route decorator which requires licences_list.is_enterprise() to be True.
    If not return a 404 http status.

    Returns:
        decorator

    """
    @functools.wraps(f)
    def decorated(*args, **kwargs):
        global licences_list
        if not licences_list.is_enterprise():
            return forbidden_no_valid_licence()

        result = f(*args, **kwargs)
        return result
    return decorated

