#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2023  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

"""
This module aims to share global variables between other modules,
so you can reload this module to update variables and avoid
restarting other modules that contains variables

->reload_globals() for initiate
"""
import os
import json
import logging

logger = logging.getLogger('globalsupdate')
hdlr = logging.StreamHandler()
hdlr.setFormatter(logging.Formatter('%(asctime)s [%(name)-15s] %(levelname)s %(message)s'))
logger.addHandler(hdlr)

def load_shared_variables():
    # wher does tasks_db_dir comes from ??? tasks.py ? how is it defined, same for conf
    path = os.path.join(tasks_db_dir, "tasks_shared_variables%s.json" % conf.get('application_root','main'))
    loaded = {}
    defaults = {}

    if not os.path.isfile(path):
        with open(path, "w", encoding="utf-8") as f:
            f.write(json.dumps(defaults))
        logger.warn('No storage file found, creating ' + path + '...')
    try:
        with open(path, "r", encoding="utf-8") as f:
            loaded = json.loads(f.read())
        defaults.update(loaded)
    except:
        logger.warn('Failed to load from storage file, registering defaults')
    globals().update(defaults)
    globals()['shared_variables'] = list(defaults.keys())

def write_shared_variables():
    path = os.path.join(tasks_db_dir, "tasks_shared_variables%s.json" % conf.get('application_root','main'))
    loaded = {}

    if 'shared_variables' not in globals().keys():
        return

    shared = globals()['shared_variables']
    for s in shared:
        loaded[s] = globals()[s]
    with open(path, "w", encoding="utf-8") as f:
        f.write(json.dumps(loaded))

def reload_protected_imports():
    from waptserver.tasks import tasks_config_file, tasks_db_dir, conf
    from waptserver.config import load_config

    globals().update(locals())

def reload_globals(allow_imports_reload=True):
    """
    Load shared variables and registering them as attributes of this module

    allow_imports_reload: Some variables needs modules to be loaded,
                          disable this if you don't want dependencies modules to be restarted

    Current shared variables:
        conf : From config file
        + all variables defined in db/tasks_shared_variables.json
    """
    try:
        logger.info('reload globals config from %s' % tasks_config_file)
        load_shared_variables()
        conf = load_config(tasks_config_file)
    except:
        if allow_imports_reload:
            reload_protected_imports()
            reload_globals(allow_imports_reload=False)

    # TODO: not very clean... ???
    globals().update(locals())

def update(var, value):
    """
    Registering or updating a shared variable
    """
    globals()[var] = value
    if var not in globals()['shared_variables']:
        globals()['shared_variables'].append(var)
    write_shared_variables()

def delete(var):
    """
    Deleting a shared variable
    """
    if var not in globals()['shared_variables']:
        return
    del globals()[var]
    globals()['shared_variables'].remove(var)
    write_shared_variables()

def force_get(var):
    """
    Return the var after reloading global variables from the storage file
    Usually you can juste acces the var as a attribute of this module (globalsupdate.var)
    """
    load_shared_variables()
    if var in globals():
        return globals()[var]
    else:
        return None
