#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2023  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import os
import datetime
import json
import requests
import time

try:
    wapt_root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
except:
    wapt_root_dir = 'c:/tranquilit/wapt'

from peewee import SQL, fn
#from waptserver.utils import *
from waptserver.utils import logger
#from waptserver.model import *
from waptserver.model import WaptDB, ServerAttribs, SyncStatus, Packages, Hosts, HostPackagesStatus, WsusUpdates, HostWsus

#from waptserver.tasks import *
from waptserver.tasks import crontab, huey

from waptserver.tasks import conf as tasks_conf

from waptutils import get_sha256

from waptserver.config import type_debian, type_redhat, type_windows
from waptserver.common import licences_list  # used as global directive below

if type_windows():
    NGINX_GID = None
else:
    try:
        import grp
        import pwd
        WAPT_UID = pwd.getpwnam('wapt').pw_uid
        if type_debian():
            NGINX_GID = grp.getgrnam('www-data').gr_gid
        elif type_redhat():
            NGINX_GID = grp.getgrnam('nginx').gr_gid
        else:
            NGINX_GID = None
    except:
        NGINX_GID = None

def change_progression_in_DB(progression=0):
    (v, created) = ServerAttribs.get_or_create(key='repo_sync_progression')
    v.value = progression
    v.save()
    (v, created) = ServerAttribs.get_or_create(key='repo_sync_in_progress')
    v.value = progression < 100
    v.save()


def get_tree_of_files_rec(adir='', all_files={}, rmpath='', changelog=None, total_files=0, total_size=0):
    for entry in os.listdir(adir):
        full_path = os.path.join(adir, entry)
        if not os.path.exists(full_path):
            continue
        minpath = os.path.normpath(os.path.relpath(full_path, rmpath)).replace(os.sep, '/')
        all_files[minpath] = {}
        if os.path.isdir(full_path):
            all_files[minpath]['isDir'] = True
            all_files[minpath]['lastmodification'] = os.path.getmtime(full_path)
            all_files[minpath]['size'] = os.path.getsize(full_path)
            if changelog:
                changelog['new'][minpath] = dict(all_files[minpath])
            all_files[minpath]['files'] = {}
            all_files[minpath]['files']['total_files_calculated'] = all_files['total_files_calculated']
            all_files[minpath]['files']['total_size_calculated'] = all_files['total_size_calculated']
            get_tree_of_files_rec(full_path, all_files[minpath]['files'], rmpath, changelog, total_files=total_files, total_size=total_size)
            all_files['total_files_calculated'] = all_files[minpath]['files']['total_files_calculated']
            all_files['total_size_calculated'] = all_files[minpath]['files']['total_size_calculated']
            del all_files[minpath]['files']['total_files_calculated']
            del all_files[minpath]['files']['total_size_calculated']
        else:
            all_files[minpath]['isDir'] = False
            all_files[minpath]['size'] = os.path.getsize(full_path)
            all_files[minpath]['lastmodification'] = os.path.getmtime(full_path)
            all_files[minpath]['sum'] = get_sha256(full_path)
            all_files['total_files_calculated'] += 1
            all_files['total_size_calculated'] += all_files[minpath]['size']
            if (all_files['total_files_calculated'] % 5 == 0) or (all_files['total_files_calculated'] == total_files):
                change_progression_in_DB(all_files['total_size_calculated'] / total_size * 100)
            if changelog:
                changelog['new'][minpath] = all_files[minpath]
    return all_files


def get_tree_of_files(dirs=[]):
    all_files = {}
    all_files_list = []
    total_files_calculated = 0
    total_size_calculated = 0
    change_progression_in_DB()
    for adir in dirs:
        for r, d, files in os.walk(adir):
            all_files_list += [os.path.join(r, afile) for afile in files]
    total_files = len(all_files_list)
    total_size = sum([os.path.getsize(afile) for afile in all_files_list])
    for adir in dirs:
        if os.path.isdir(adir):
            minpath = os.path.normpath(os.path.relpath(adir, os.path.dirname(adir))).replace(os.sep, '/')
            all_files[minpath] = {}
            all_files[minpath]['lastmodification'] = os.path.getmtime(adir)
            all_files[minpath]['size'] = os.path.getsize(adir)
            all_files[minpath]['realpath'] = os.path.normpath(adir)
            all_files[minpath]['isDir'] = True
            all_files[minpath]['files'] = {}
            all_files[minpath]['files']['total_files_calculated'] = total_files_calculated
            all_files[minpath]['files']['total_size_calculated'] = total_size_calculated
            get_tree_of_files_rec(adir, all_files[minpath]['files'], os.path.dirname(adir), total_files=total_files, total_size=total_size)
            total_files_calculated = all_files[minpath]['files']['total_files_calculated']
            total_size_calculated = all_files[minpath]['files']['total_size_calculated']
            del all_files[minpath]['files']['total_files_calculated']
            del all_files[minpath]['files']['total_size_calculated']
    return all_files


def put_tree_of_files_in_sync_file(filesync='', dirs=[], username='SERVER'):
    time_start = time.time()
    with open(filesync, 'w') as f:
        try:
            id = int(SyncStatus.select(fn.MAX(SyncStatus.id)).scalar())+1
        except:
            id = 1
        SyncStatus.delete().execute()
        ServerAttribs.delete_by_id('repo_sync_progression')
        tree_of_files = get_tree_of_files(dirs)
        changelog = {}
        changelog['new'] = tree_of_files
        data = {'updated_by': username, 'updated_on': datetime.datetime.utcnow(), 'created_by': username, 'created_on': datetime.datetime.utcnow(), 'version': 1, 'changelog': changelog}
        changelog['time spent'] = time.time()-time_start
        SyncStatus.create(**data)
        tree_of_files_version = {}
        tree_of_files_version['version'] = 1
        tree_of_files_version['id'] = id
        tree_of_files_version['files'] = tree_of_files
        json.dump(tree_of_files_version, f, indent=4)
        return tree_of_files_version


def actualize_tree_of_files_rec(adir='', dico={}, rmpath='', changelog=None, total_files=0, total_size=0):
    listdirdico = list(dico.keys())
    for entry in os.listdir(adir):
        fullpath = os.path.join(adir, entry)
        minpath = os.path.normpath(os.path.relpath(fullpath, rmpath)).replace(os.sep, '/')
        if (dico.get(minpath)):
            listdirdico.remove(minpath)
            if (dico[minpath]['lastmodification'] == os.path.getmtime(fullpath)) and (dico[minpath]['size'] == os.path.getsize(fullpath)):
                if dico[minpath]['isDir']:
                    dico[minpath]['files']['total_files_calculated'] = dico['total_files_calculated']
                    dico[minpath]['files']['total_size_calculated'] = dico['total_size_calculated']
                    actualize_tree_of_files_rec(fullpath, dico[minpath]['files'], rmpath, changelog, total_files, total_size=total_size)
                    dico['total_files_calculated'] = dico[minpath]['files']['total_files_calculated']
                    dico['total_size_calculated'] = dico[minpath]['files']['total_size_calculated']
                    del dico[minpath]['files']['total_files_calculated']
                    del dico[minpath]['files']['total_size_calculated']
                else:
                    dico['total_files_calculated'] += 1
                    dico['total_size_calculated'] += dico[minpath]['size']
                    if (dico['total_files_calculated'] % 5 == 0) or (dico['total_files_calculated'] == total_files):
                        change_progression_in_DB(dico['total_size_calculated'] / total_size * 100)
                    continue
            else:
                if os.path.isdir(fullpath):
                    dico[minpath]['lastmodification'] = os.path.getmtime(fullpath)
                    dico[minpath]['size'] = os.path.getsize(fullpath)
                    changelog['modified'][minpath] = dict(dico[minpath])
                    if (dico[minpath]['isDir']):
                        dico[minpath]['files']['total_files_calculated'] = dico['total_files_calculated']
                        dico[minpath]['files']['total_size_calculated'] = dico['total_size_calculated']
                        actualize_tree_of_files_rec(fullpath, dico[minpath]['files'], rmpath, changelog, total_files, total_size=total_size)
                        dico['total_files_calculated'] = dico[minpath]['files']['total_files_calculated']
                        dico['total_size_calculated'] = dico[minpath]['files']['total_size_calculated']
                        del dico[minpath]['files']['total_files_calculated']
                        del dico[minpath]['files']['total_size_calculated']
                    else:
                        dico[minpath]['isDir'] = True
                        dico[minpath]['files'] = {}
                        del dico[minpath]['sum']
                        dico[minpath]['files']['total_files_calculated'] = dico['total_files_calculated']
                        dico[minpath]['files']['total_size_calculated'] = dico['total_size_calculated']
                        get_tree_of_files_rec(fullpath, dico[minpath]['files'], rmpath, changelog, total_files, total_size=total_size)
                        dico['total_files_calculated'] = dico[minpath]['files']['total_files_calculated']
                        dico['total_size_calculated'] = dico[minpath]['files']['total_size_calculated']
                        del dico[minpath]['files']['total_files_calculated']
                        del dico[minpath]['files']['total_size_calculated']
                    del changelog['modified'][minpath]['files']
                else:
                    dico[minpath]['lastmodification'] = os.path.getmtime(fullpath)
                    dico[minpath]['sum'] = get_sha256(fullpath)
                    dico[minpath]['size'] = os.path.getsize(fullpath)
                    if (dico[minpath]['isDir']):
                        dico[minpath]['isDir'] = False
                        del dico[minpath]['files']
                    changelog['modified'][minpath] = dico[minpath]
                    dico['total_files_calculated'] += 1
                    if (dico['total_files_calculated'] % 5 == 0) or (dico['total_files_calculated'] == total_files):
                        change_progression_in_DB(dico['total_size_calculated'] / total_size * 100)
        elif os.path.exists(fullpath):
            dico[minpath] = {}
            dico[minpath]['lastmodification'] = os.path.getmtime(fullpath)
            dico[minpath]['size'] = os.path.getsize(fullpath)
            if os.path.isdir(fullpath):
                dico[minpath]['isDir'] = True
                changelog['new'][minpath] = dict(dico[minpath])
                dico[minpath]['files'] = {}
                dico[minpath]['files']['total_files_calculated'] = dico['total_files_calculated']
                dico[minpath]['files']['total_size_calculated'] = dico['total_size_calculated']
                get_tree_of_files_rec(fullpath, dico[minpath]['files'], rmpath, changelog, total_files, total_size=total_size)
                dico['total_files_calculated'] = dico[minpath]['files']['total_files_calculated']
                dico['total_files_calculated'] = dico[minpath]['files']['total_size_calculated']
                del dico[minpath]['files']['total_files_calculated']
                del dico[minpath]['files']['total_size_calculated']
            else:
                dico[minpath]['isDir'] = False
                dico[minpath]['sum'] = get_sha256(fullpath)
                changelog['new'][minpath] = dico[minpath]
                dico['total_files_calculated'] += 1
                if (dico['total_files_calculated'] % 5 == 0) or (dico['total_files_calculated'] == total_files):
                    change_progression_in_DB(dico['total_size_calculated'] / total_size * 100)
    for entry in listdirdico:
        if entry!='total_files_calculated' and entry!='total_size_calculated':
            changelog['deleted'][entry] = dict(dico[entry])
            if dico[entry]['isDir']:
                del changelog['deleted'][entry]['files']
            del dico[entry]


def actualize_tree_of_files(dico={}, changelog={}):
    changelog['new'] = {}
    changelog['deleted'] = {}
    changelog['modified'] = {}
    all_files_list = []
    total_files_calculated = 0
    total_size_calculated = 0
    change_progression_in_DB()
    dirs = [tasks_conf['wapt_folder'], tasks_conf['waptwua_folder'], tasks_conf['wapt_folder']+'-host' , tasks_conf['wads_folder']]
    for f in dirs:
        folder = f.split('/')[-1].split('\\')[-1]
        if not folder in dico.keys():
            dico[folder] = {"isDir":True,"realpath":f,"lastmodification":os.path.getmtime(f),"size":os.path.getsize(f),"files":{}}
    for adir in dico.keys():
        for r, d, files in os.walk(dico[adir]['realpath']):
            all_files_list += [os.path.join(r,afile) for afile in files]
    total_files = len(all_files_list)
    total_size = sum([os.path.getsize(afile) for afile in all_files_list])
    for adir in dico:
        dico[adir]['files']['total_files_calculated'] = total_files_calculated
        dico[adir]['files']['total_size_calculated'] = total_size_calculated
        actualize_tree_of_files_rec(dico[adir]['realpath'], dico[adir]['files'], os.path.dirname(dico[adir]['realpath']), changelog, total_files, total_size)
        total_files_calculated = dico[adir]['files']['total_files_calculated']
        total_size_calculated = dico[adir]['files']['total_size_calculated']
        del dico[adir]['files']['total_files_calculated']
        del dico[adir]['files']['total_size_calculated']
    change_progression_in_DB(100)
    return dico


def check_renamed(changelog):
    remove_from_new_file = []
    for new_file in changelog['new'].keys():
        found = False
        remove_from_deleted_file = []
        if not(changelog['new'][new_file]['isDir']):
            for deleted_file in changelog['deleted'].keys():
                if not(changelog['deleted'][deleted_file]['isDir']) and (changelog['new'][new_file]['sum'] == changelog['deleted'][deleted_file]['sum']):
                    if 'moved/renamed' not in changelog:
                        changelog['moved/renamed'] = {}
                    changelog['moved/renamed'][new_file] = changelog['deleted'][deleted_file]
                    changelog['moved/renamed'][new_file]['old_location'] = deleted_file
                    remove_from_deleted_file.append(deleted_file)
                    found = True
        if found:
            remove_from_new_file.append(new_file)
    for deleted_file in remove_from_deleted_file:
        del changelog['deleted'][deleted_file]
    for new_file in remove_from_new_file:
        del changelog['new'][new_file]


def actualize_tree_of_files_in_sync_file(filesync='', username='SERVER'):
    time_start = time.time()
    with open(filesync, 'r+') as f:
        tree_of_files = json.load(f)
        changelog = {}
        new_tree = {}
        new_tree['files'] = actualize_tree_of_files(tree_of_files['files'], changelog)
        if changelog['new'] or changelog['deleted'] or changelog['modified']:
            if changelog['new']:
                check_renamed(changelog)
            new_tree['version'] = tree_of_files['version']+1
            new_tree['id'] = tree_of_files['id']
            data = {'updated_by': username, 'updated_on': datetime.datetime.utcnow(), 'created_by': username, 'created_on': datetime.datetime.utcnow(), 'version': new_tree['version'], 'changelog': changelog}
            changelog['time spent'] = time.time()-time_start
            SyncStatus.create(**data)
            f.truncate(0)
            f.seek(0)
            json.dump(new_tree, f, indent=4)
            return new_tree
        else:
            return tree_of_files


def update_differenciated_index_repositories(inadir=None):
    if inadir is None:
        inadir = tasks_conf['wapt_folder'] + '-diff-repos'

    if not(os.path.isdir(inadir)):
        os.mkdir(inadir)
        if NGINX_GID is not None:
            os.chown(inadir, WAPT_UID, NGINX_GID)

    list_agent_repo = list(Hosts.select(Hosts.uuid, SQL("wapt_status ->> 'remote_repo_url' AS remote_repo_url")).where(Hosts.wapt_status.contains({'is_remote_repo': True})).dicts())

    updated = False

    for agent_repo in list_agent_repo:

        where_clause = (fn.CONCAT(';', Hosts.repositories, ';').contains(';'+agent_repo['remote_repo_url']+';'))
        full_list_of_packages_error = list(Hosts.select(Packages.filename).distinct().where(where_clause & (HostPackagesStatus.install_status == 'ERROR')).join(HostPackagesStatus, on=(Hosts.uuid == HostPackagesStatus.host_id)).join(Packages, on=(Packages.package_uuid == HostPackagesStatus.package_uuid)).dicts())
        full_list_of_packages_need_install = list(Hosts.select(Packages.filename).distinct().where(where_clause & (HostPackagesStatus.install_status == 'NEED-INSTALL')).join(HostPackagesStatus, on=(Hosts.uuid == HostPackagesStatus.host_id)).join(Packages, on=((Packages.version == HostPackagesStatus.version) & (Packages.package == HostPackagesStatus.package))).dicts())
        full_list_of_cabs = list(Hosts.select(WsusUpdates.download_urls.alias('urls')).distinct().where(where_clause & (HostWsus.allowed == True) & (HostWsus.installed == False) & (HostWsus.downloaded == False)).join(HostWsus, on=(Hosts.uuid == HostWsus.host_id)).join(WsusUpdates, on=(HostWsus.update_id == WsusUpdates.update_id)).dicts())
        full_list_of_packages_host = list(Hosts.select(fn.CONCAT(Hosts.uuid, '.wapt').alias('filename')).distinct().where(where_clause & (HostPackagesStatus.section == 'host')).join(HostPackagesStatus, on=(Hosts.uuid == HostPackagesStatus.host_id)).dicts())
        file_agent_repo = os.path.join(inadir, agent_repo['uuid']+'.json')

        if not os.path.isfile(file_agent_repo):
            with open(file_agent_repo, 'w'):
                pass
            if NGINX_GID is not None:
                os.chown(file_agent_repo, WAPT_UID, NGINX_GID)

        if full_list_of_packages_error or full_list_of_packages_need_install or full_list_of_packages_host or full_list_of_cabs:
            updated = True

        with open(file_agent_repo, 'r+') as agent_repo_file:
            try:
                json_agent_repo = json.load(agent_repo_file)
                agent_repo_file.truncate(0)
                agent_repo_file.seek(0)
            except:
                json_agent_repo = {
                  'files': {
                    'wapt':
                      {'manual_add': [], 'auto_add': ['Packages', 'waptagent.exe']},
                    'waptagent':
                      {'manual_add': [], 'auto_add': ['waptsetup.exe', 'waptdeploy.exe']},
                    'waptwua':
                      {'manual_add': [], 'auto_add': ['wsusscn2.cab']}, 'wapt-host': {'manual_add': [], 'auto_add': []}
                    },
                  'version': 0}
            json_agent_repo['files']['wapt']['auto_add'] += [elem['filename'] for elem in full_list_of_packages_need_install+full_list_of_packages_error if elem['filename'] not in json_agent_repo['files']['wapt']['auto_add']]
            for alist_of_cabs in full_list_of_cabs:
                for aurl in alist_of_cabs['urls']:
                    filename = aurl[aurl.rfind('/')+1:]
                    if filename not in json_agent_repo['files']['waptwua']['auto_add']:
                        json_agent_repo['files']['waptwua']['auto_add'].append(filename)
            json_agent_repo['files']['wapt-host']['auto_add'] += [elem['filename'] for elem in full_list_of_packages_host if elem['filename'] not in json_agent_repo['files']['wapt-host']['auto_add']]
            json_agent_repo['version'] += 1
            json.dump(json_agent_repo, agent_repo_file, indent=4)

    return updated

@huey.periodic_task(crontab(hour='*/'+str(tasks_conf['remote_repo_update_delay']), minute='0'))
@huey.lock_task('tasks_sync_json')
def update_file_tree_of_files_in_huey(filesync=None, dirs=None, username='SERVER', emit_websocket=None, force=False, task=None):
    if not tasks_conf['repo_sync_enable']:
        return 'NOT ENABLED'

    def launch_emit_on_server():
        r = requests.get('http://127.0.0.1:%s/api/v3/broadcast_sync_on_remotes_repo' % tasks_conf['waptserver_port'])
        if r.status_code == 200:
            r = r.json()
            if r['success']:
                logger.info(r['msg'])
            else:
                logger.info('ERROR: %s' % r['msg'])
        else:
            logger.info('Connexion impossible to localhost through http for emitting sync message with websocket')

    global licences_list
    if filesync is None:
        filesync = os.path.join(os.path.abspath(os.path.join(tasks_conf['wapt_folder'], os.pardir)), 'sync.json')
    if dirs is None:
        dirs = [ tasks_conf['wapt_folder'], tasks_conf['waptwua_folder'], tasks_conf['wapt_folder']+'-host' , tasks_conf['wads_folder']]
    if emit_websocket is None:
        emit_websocket=tasks_conf.get('remote_repo_websockets')
    if licences_list.is_enterprise():
        with WaptDB():
            if Hosts.select().where(Hosts.wapt_status.contains({'is_remote_repo':True})).count():
                have_to_emit_websocket = False
                if (not(os.path.isfile(filesync)) or (os.stat(filesync).st_size == 0)) or (not(SyncStatus.select(SyncStatus.version)).exists()):
                    put_tree_of_files_in_sync_file(filesync, dirs, username)
                    have_to_emit_websocket = emit_websocket
                else:
                    with open(filesync, 'r') as f:
                        version = json.load(f)['version']
                    res = actualize_tree_of_files_in_sync_file(filesync, username)
                    have_to_emit_websocket = emit_websocket and res['version'] > version

                if tasks_conf.get('remote_repo_diff'):
                    list_agent_repo = list(Hosts.select(SQL("wapt_status ->> 'remote_repo_url' AS remote_repo_url")).where(Hosts.wapt_status.contains({'is_remote_repo': True})).dicts())
                    where_clause = (fn.CONCAT(';', Hosts.repositories, ';') << [';'+agent_repo['remote_repo_url']+';' for agent_repo in list_agent_repo])

                    if not(force):
                        number_of_package_missing = Hosts.select(Hosts.uuid).distinct().where(where_clause & ((HostPackagesStatus.install_status == 'NEED-INSTALL') | (HostPackagesStatus == 'ERROR'))).join(HostPackagesStatus, on=(Hosts.uuid == HostPackagesStatus.host_id)).count()
                        number_of_cabs_missing = Hosts.select(Hosts.uuid).distinct().where(where_clause & (HostWsus.allowed == True) & (HostWsus.installed == False) & (HostWsus.downloaded == False)).join(HostWsus, on=(Hosts.uuid == HostWsus.host_id)).count()

                    if force or number_of_package_missing or number_of_cabs_missing:
                        updated_diff = update_differenciated_index_repositories()
                        have_to_emit_websocket = emit_websocket and (updated_diff or have_to_emit_websocket)

                if have_to_emit_websocket:
                    launch_emit_on_server()
    else:
        return 'NOT ENTERPRISE'

if __name__ == '__main__':
    update_file_tree_of_files_in_huey()
