#!/usr/bin/env python
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2023  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

import sys
import logging
import os
import socket
import pefile
import errno
import hashlib

if sys.platform == 'win32':
    try:
        # it does not matter if this fails.
        # it would mean that calling process has already initialized it.
        import pythoncom
        pythoncom.CoInitializeEx(pythoncom.COINIT_MULTITHREADED)
    except:
        pass
    import wmi

__all__ = [
    'mkdir_p',
    'get_disk_space',
    'logger',
    'setloglevel',
    'get_dns_domain',
    'get_wapt_exe_version',
    'get_hash256_exe',
    'EWaptMissingHostData',
    'EWaptUnknownHost',
    'EWaptHostUnreachable',
    'EWaptForbiddden',
    'EWaptMissingParameter',
    'EWaptSignalReceived',
    'EWaptDatabaseError',
    'EWaptAuthenticationFailure',
    'EWaptTimeoutWaitingForResult',
]

logger = logging.getLogger('waptserver')

def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def get_disk_space(directory):
    ret = None

    if os.name == 'posix':
        stats = os.statvfs(directory)  # pylint: disable=no-member
        ret = (stats.f_bavail * stats.f_bsize, stats.f_blocks * stats.f_bsize)
    else:
        drive = os.path.splitdrive(os.path.abspath(directory))[0].lower()

        for d in wmi.WMI().Win32_LogicalDisk():
            if str(d.Name).lower() == drive:
                ret = (int(d.FreeSpace), int(d.Size))

    return ret


def setloglevel(logger, loglevel):
    """set loglevel as string"""
    if loglevel in ('debug', 'warning', 'info', 'error', 'critical'):
        numeric_level = getattr(logging, loglevel.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: {}'.format(loglevel))
        logger.setLevel(numeric_level)


def get_dns_domain():
    """Get DNS domain part of the FQDN

    Returns:
        str
    """
    try:
        parts = socket.getfqdn().lower().split('.', 1)
        if len(parts) > 1:
            return parts[1]
        else:
            return ''
    except Exception as e:
        logger.critical('Unable to get DNS domain: %s' % e)
        return None



def get_wapt_exe_version(exe):
    """Returns FileVersion or ProductVersion of windows executables

    Args:
        exe (str): path to executable

    Returns:
        (bool,str): (is present, agent version)
    """

    present = False
    version = None
    if os.path.exists(exe):
        present = True
        pe = None
        try:
            pe = pefile.PE(exe,fast_load=True)
            pe.parse_data_directories([pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_RESOURCE']])
            version = pe.FileInfo[0][0].StringTable[
                0].entries[b'FileVersion'].strip().decode("utf-8")
            if not version:
                version = pe.FileInfo[0][0].StringTable[
                    0].entries[b'ProductVersion'].strip().decode("utf-8")
        except:
            pass
        if pe is not None:
            pe.close()
    return (present, version)

def get_hash256_exe(exe_filename):
    BUF_SIZE = 65536

    if os.path.isfile(exe_filename):
        hash_exe = hashlib.sha256()
        with open(exe_filename, 'rb') as exe:
            while True:
                data_exe = exe.read(BUF_SIZE)
                if not data_exe:
                    break
                hash_exe.update(data_exe)
        return hash_exe.hexdigest()
    else:
        return ''

def get_forwarded_url(request):
    proto = request.headers.get('X-Forwarded-Proto','http')
    host_url = request.host_url.split('//')[1]
    return '%s://%s' % (proto,host_url)

def get_all_json_configs(conf_d_path):
    """Returns a list of all available dynamic configurations.
    content is returned as a str representation, as stored in file.

    Returns
        list of dict {filename,name,hash,content}
    """
    configs = []
    for file in os.listdir(conf_d_path):
        file_abspath = os.path.join(conf_d_path, file)
        if not os.path.isfile(file_abspath) or not file.endswith('.json'):
            continue
        try:
            if len(file) >= 70 and  file[-70] == '_':
                config = {"name":file[:-70], "hash":file[-69:-5], "filename":file}
            else:
                config = {"name":file[:-5], "hash":'', "filename":file}
            with open(file_abspath, 'r') as config_file:
                config["content"] = config_file.read()
            configs.append(config)
        except:
            continue
    return configs


# Custom exceptions #####
class EWaptMissingHostData(Exception):
    pass


class EWaptUnknownHost(Exception):
    pass


class EWaptHostUnreachable(Exception):
    pass


class EWaptForbiddden(Exception):
    pass


class EWaptMissingParameter(Exception):
    pass


class EWaptSignalReceived(Exception):
    pass


class EWaptDatabaseError(Exception):
    pass


class EWaptAuthenticationFailure(Exception):
    pass


class EWaptTimeoutWaitingForResult(Exception):
    pass
