#!/opt/wapt/python
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2023  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

from waptutils import __version__

import shutil
import os
import sys

if "__file__" in locals():
    sys.path.insert(0, os.path.realpath(os.path.join( os.path.dirname(os.path.realpath(__file__)),'..')))

from win32api import GetUserNameEx,NameSamCompatible

try:
    wapt_root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
except:
    wapt_root_dir = 'c:/tranquilit/wapt'

from optparse import OptionParser
import logging
import setuphelpers
import datetime

import jinja2
import time
import random
import string
import base64
import uuid
import socket
import traceback
import glob
import hashlib
import json
import win32api

from setuphelpers import run
from waptutils import setloglevel, ensure_unicode, get_local_IPs, makepath
from waptcrypto import SSLPrivateKey, SSLCertificate, SSLCRL
from passlib.hash import pbkdf2_sha256

import waptserver.config
from waptserver.utils import mkdir_p
from waptserver.common import update_all_crls

from waptpackage import WaptLocalRepo
from waptserver.model import load_db_config, Packages

logger = logging.getLogger('waptwinsetup')

def system32():
    return win32api.GetSystemDirectory()

def fqdn():
    """ Guess FQDN from hostname - return default value if nothing found"""
    try:
        fqdn = socket.getfqdn()
        if not fqdn:
            fqdn = 'wapt'
        if '.' not in fqdn:
            fqdn += '.lan'
    except:
        fqdn = 'srvwapt'
    return fqdn

def create_dhparam(key_size=2048):
    logger.info('Create dhparam size %s. This can take a while...' % key_size)
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives.asymmetric import dh
    parameters = dh.generate_parameters(generator=2, key_size=key_size, backend=default_backend())
    return parameters.parameter_bytes(serialization.Encoding.PEM, format=serialization.ParameterFormat.PKCS3)


def make_nginx_config(wapt_root_dir, conf, force=False, server_names=None):
    """Create a nginx default config file to server wapt_folder and reverse proxy waptserver
    Create a key and self signed certificate.

    Args:
        wapt_root_dir (str)
        conf (dict) : waptserver config file

    Returns:
        str: path to nginx conf file
    """
    ap_conf_dir = os.path.join(
        wapt_root_dir,
        'waptserver',
        'nginx',
        'conf')
    ap_file_name = 'nginx.conf'
    ap_conf_file = os.path.join(ap_conf_dir, ap_file_name)
    ap_ssl_dir = os.path.join(wapt_root_dir, 'waptserver', 'nginx', 'ssl')

    ap_ssl_dhparam_file = os.path.join(ap_ssl_dir, 'dhparam.pem')
    if not (os.path.isdir(ap_ssl_dir)):
        os.makedirs(ap_ssl_dir)
    if not(os.path.isfile(ap_ssl_dhparam_file)) or os.stat(ap_ssl_dhparam_file).st_size==0:
        with open(ap_ssl_dhparam_file, 'wb') as f:
            f.write(create_dhparam(dhparam_key_size))

    if os.path.isfile(ap_conf_file) and not force:
        with open(ap_conf_file, 'r') as f:
            if 'waptserver' in f.read():
                return ap_conf_file

    setuphelpers.mkdirs(ap_ssl_dir)

    key_fn = os.path.join(ap_ssl_dir, 'key.pem')
    key = SSLPrivateKey(key_fn)
    if not os.path.isfile(key_fn):
        logger.info('Create SSL RSA Key %s' % key_fn)
        key.create()
        key.save_as_pem()

    cert_fn = os.path.join(ap_ssl_dir, 'cert.pem')

    ipaddresses = [i for i in get_local_IPs() if not i.startswith('fe80::')]
    server_names = server_names or []

    if server_names:
        cn = server_names[0]
    else:
        cn = fqdn()

    if server_names:
        altnames = server_names
    else:
        altnames = [cn,ipaddresses[0]]

    if os.path.isfile(cert_fn):
        crt = SSLCertificate(cert_fn)
        if crt.cn != cn:
            shutil.move(cert_fn, "%s-%s.old" % (cert_fn, '{:%Y%m%d-%Hh%Mm%Ss}'.format(datetime.datetime.utcnow())))
            crt = key.build_sign_certificate(
                cn=cn,
                altnames=altnames,
                is_server_auth=True,
                is_client_auth=False,
                is_code_signing=False,
                is_ca=False,
                key_usages=['digital_signature']
                )
            logger.info('Create X509 cert %s' % cert_fn)
            crt.save_as_pem(cert_fn)
    else:
        crt = key.build_sign_certificate(
            cn=cn,
            altnames=altnames,
            is_server_auth=True,
            is_client_auth=False,
            is_code_signing=False,
            is_ca=False,
            key_usages=['digital_signature']
            )
        logger.info('Create X509 cert %s' % cert_fn)
        crt.save_as_pem(cert_fn)

    # write config file
    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(os.path.join(wapt_root_dir, 'waptserver', 'scripts')))
    template = jinja_env.get_template('waptwindows.nginxconfig.j2')
    template_variables = {
        'server_names': server_names,
        'wapt_repository_path': os.path.dirname(conf['wapt_folder']).replace('\\', '/'),
        'wapt_folder': os.path.dirname(conf['wapt_folder']).replace('\\', '/'),
        'waptserver_port': conf['waptserver_port'],
        'windows': True,
        'ssl': True,
        'force_https': False,
        'use_kerberos': False,
        'use_uwsgi': False,
        'wapt_ssl_key_file': key_fn.replace('\\', '/'),
        'wapt_ssl_cert_file': cert_fn.replace('\\', '/'),
        'log_dir': os.path.join(wapt_root_dir, 'log').replace('\\', '/'),
        'wapt_root_dir': wapt_root_dir.replace('\\', '/'),
        'nginx_http': conf['nginx_http'],
        'nginx_https': conf['nginx_https'],
        'use_ssl_client_auth': conf.get('use_ssl_client_auth', False),
        'known_certificates_folder': conf.get('known_certificates_folder', None) and conf.get('known_certificates_folder', None).replace('\\', '/'),
        'clients_signing_certificate': conf.get('clients_signing_certificate') and conf.get('clients_signing_certificate').replace('\\', '/'),
        'clients_signing_crl': conf.get('clients_signing_crl', None) and conf.get('clients_signing_crl', None).replace('\\', '/'),
        'ssl_client_crls': conf.get('ssl_client_crls', None) and conf.get('ssl_client_crls', None).replace('\\', '/'),
        'ssl_client_certificates': conf.get('ssl_client_certificates', None) and conf.get('ssl_client_certificates', None).replace('\\', '/'),
        'htpasswd_path': conf.get('htpasswd_path', None) and conf.get('htpasswd_path', None).replace('\\', '/'),
        'wapt_dhparam_file': ap_ssl_dhparam_file.replace('\\', '/'),
        'has_dhparam': os.path.isfile(ap_ssl_dhparam_file),
        'has_ipv6':waptserver.config.is_server_ipv6_capable(),
        'wads_enable': conf.get('wads_enable'),
        'waptwua_enable': conf.get('waptwua_enable'),
        'trusted_wads_subnets': conf.get('trusted_wads_subnets'),
        'login_on_wads': conf.get('login_on_wads'),
        'list_subnet_skip_login_wads': conf.get('list_subnet_skip_login_wads'),
        'homepage_enable' : conf.get('homepage_enable',True)
    }

    config_string = template.render(template_variables)
    logger.info('Create nginx conf file %s' % ap_conf_file)
    with open(ap_conf_file, 'wt') as dst_file:
        dst_file.write(config_string)
    return ap_conf_file


def install_nginx_service(options, conf=None, server_names=None, ipaddresses=None):
    # remove old nginx service
    service_name = 'WAPTNginx'
    if setuphelpers.service_installed(service_name):
        if setuphelpers.service_is_running(service_name):
            setuphelpers.service_stop(service_name)
        setuphelpers.service_delete(service_name)
        time.sleep(2)

    # nginx process is managed with waptserver service
    if setuphelpers.service_installed('WAPTServer'):
        if setuphelpers.service_is_running('WAPTServer'):
            setuphelpers.service_stop('WAPTServer')

    if conf is None:
        conf = waptserver.config.load_config(options.configfile)

    logger.info("Register nginx frontend")
    repository_path = os.path.join(wapt_root_dir, 'waptserver', 'repository')
    icacls_path = makepath(system32(), 'icacls.exe')
    for repo_path in ('.', 'wapt', 'wapt-host', 'waptwua'):
        mkdir_p(os.path.join(repository_path, repo_path))
        run(r'"%s" "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % (icacls_path, os.path.join(repository_path, repo_path)))
    mkdir_p(os.path.join(wapt_root_dir, 'waptserver', 'nginx', 'temp'))
    run(r'"%s" "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % (icacls_path, os.path.join(wapt_root_dir, 'waptserver', 'nginx', 'temp')))

    run(r'"%s" "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % (icacls_path, os.path.join(wapt_root_dir, 'waptserver', 'nginx', 'logs')))
    run(r'"%s" "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % (icacls_path, os.path.join(wapt_root_dir, 'log')))

    make_nginx_config(wapt_root_dir, conf, force=options.force, server_names=server_names)

    update_all_crls(conf)


def migrate_pg_db(old_pgsql_root_dir, old_pgsql_data_dir, pgsql_root_dir, pgsql_data_dir):
    #
    try:
        cwd = os.getcwd()
        tmpdir = os.path.join(cwd, 'temp')
        mkdir_p(tmpdir)
        os.chdir(tmpdir)
        icacls_path = makepath(system32(), 'icacls.exe')
        setuphelpers.run(r'"%s" "%s" /t /grant  "*S-1-1-0":(OI)(CI)(M)' % (icacls_path, tmpdir))

        # force right for current user
        setuphelpers.run(r'"%s" "%s" /grant "%s":(OI)(CI)F /t' % (icacls_path, pgsql_data_dir,GetUserNameEx(NameSamCompatible)))
        setuphelpers.run(r'"%s" "%s" /grant "%s":(OI)(CI)F /t' % (icacls_path, old_pgsql_data_dir,GetUserNameEx(NameSamCompatible)))

        old_pgsql_root_dir = old_pgsql_root_dir.replace('\\', '/')
        old_pgsql_data_dir = old_pgsql_data_dir.replace('\\', '/')
        pgsql_root_dir = pgsql_root_dir.replace('\\', '/')
        pgsql_data_dir = pgsql_data_dir.replace('\\', '/')

        cmd = r'"{pgsql_root_dir}/bin/pg_upgrade.exe" -U postgres --old-datadir "{old_pgsql_data_dir}" --new-datadir "{pgsql_data_dir}" --old-bindir "{old_pgsql_root_dir}/bin" --new-bindir "{pgsql_root_dir}/bin"'.format(**locals())

        logger.info('Running %s' % cmd)
        logger.info(run(cmd, cwd=tmpdir))


        #reforce acl
        setuphelpers.run(r'"%s" "%s" /setowner "%s" /T' % (icacls_path, pgsql_data_dir,GetUserNameEx(NameSamCompatible)))
        setuphelpers.run(r'"%s" "%s" /t /grant  "%s":(OI)(CI)F /T' % (icacls_path, pgsql_data_dir, GetUserNameEx(NameSamCompatible)))
        setuphelpers.run(r'"%s" "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % (icacls_path, log_directory))
        setuphelpers.run(r'"%s" "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % (icacls_path, pgsql_data_dir))

        shutil.move(old_pgsql_root_dir, old_pgsql_root_dir+'.old')
        shutil.move(old_pgsql_data_dir, old_pgsql_data_dir+'.old')
        return True
    except Exception as e:
        logger.info('Unable to migrate database : %s' % ensure_unicode(e))
        return False
    finally:
        os.chdir(cwd)


def configure_max_connections(pg_sql_data_dir):
    config_file_path = os.path.join(pg_sql_data_dir, 'postgresql.conf')
    with open(config_file_path, "r+") as config_file:
        new_lines = []
        found = False
        for line in config_file.readlines():
            if 'max_connections' in line:
                new_lines.append('max_connections = 1000\n')
                found = True
            else:
                new_lines.append(line)
        if not(found):
            new_lines.append('max_connections = 1000\n')
        config_file.seek(0)
        config_file.truncate()
        config_file.writelines(new_lines)


def install_postgresql_service(options, conf=None):
    if conf is None:
        conf = waptserver.config.load_config(options.configfile)
    logger.info("Install postgres database")

    pgsql_root_dir = r'%s\waptserver\pgsql-14' % wapt_root_dir
    pgsql_data_dir = r'%s\waptserver\pgsql_data-14' % wapt_root_dir
    pgsql_data_dir = pgsql_data_dir.replace('\\', '/')

    logger.info("About to build database directory")
    if setuphelpers.service_installed('waptserver') and setuphelpers.service_is_running('waptserver'):
        logger.info('Stopping waptserver')
        setuphelpers.service_stop('waptserver')
        # waiting for postgres to be ready
        time.sleep(1)

    if setuphelpers.service_installed('wapttasks') and setuphelpers.service_is_running('wapttasks'):
        logger.info('Stopping wapttasks')
        setuphelpers.service_stop('wapttasks')
        # waiting for postgres to be ready
        time.sleep(1)

    if setuphelpers.service_installed('waptpostgresql') and setuphelpers.service_is_running('waptpostgresql'):
        logger.info('Stopping postgresql')
        setuphelpers.service_stop('waptpostgresql')
        # waiting for postgres to be ready
        time.sleep(2)

    if not os.path.exists(os.path.join(pgsql_data_dir, 'postgresql.conf')):
        setuphelpers.mkdirs(pgsql_data_dir)

    icacls_path = makepath(system32(), 'icacls.exe')
    # need to have specific write acls for current user otherwise initdb fails...
    setuphelpers.run(r'"%s" "%s" /setowner "%s" /T' % (icacls_path, pgsql_data_dir,GetUserNameEx(NameSamCompatible)))
    setuphelpers.run(r'"%s" "%s" /t /grant  "%s":(OI)(CI)F /T' % (icacls_path, pgsql_data_dir, GetUserNameEx(NameSamCompatible)))
    setuphelpers.run(r'"%s" "%s" /t /grant  "*S-1-5-20":(OI)(CI)(M) /T' % (icacls_path, pgsql_data_dir))

        #setuphelpers.run(r'"%s" "%s" /inheritance:r' % (icacls_path, pgsql_data_dir,))
    if not os.path.exists(os.path.join(pgsql_data_dir, 'postgresql.conf')):
        setuphelpers.run(r'"%s\bin\initdb" -U postgres -E UTF8 -D "%s"' % (pgsql_root_dir, pgsql_data_dir))
    else:
        logger.info("Database server data dir already instanciated, doing nothing")

    logger.info("Start postgresql database")

    if setuphelpers.service_installed('WaptPostgresql'):
        if setuphelpers.service_is_running('WaptPostgresql'):
            setuphelpers.service_stop('waptPostgresql')
        setuphelpers.service_delete('waptPostgresql')

    cmd = r'"%s\bin\pg_ctl" register -N WAPTPostgresql -U "nt authority\networkservice" -S auto -D "%s"  ' % (pgsql_root_dir, pgsql_data_dir)
    run(cmd)

    #reforce acl
    setuphelpers.run(r'"%s" "%s" /setowner "%s" /T' % (icacls_path, pgsql_data_dir,GetUserNameEx(NameSamCompatible)))
    setuphelpers.run(r'"%s" "%s" /t /grant  "%s":(OI)(CI)F /T' % (icacls_path, pgsql_data_dir, GetUserNameEx(NameSamCompatible)))
    setuphelpers.run(r'"%s" "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % (icacls_path, log_directory))
    setuphelpers.run(r'"%s" "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % (icacls_path, pgsql_data_dir))

    # try to migrate from old version (pg 9.6 to pg 14)
    old_pgsql_root_dir = r'%s\waptserver\pgsql-9.6' % wapt_root_dir
    old_pgsql_data_dir = r'%s\waptserver\pgsql_data-9.6' % wapt_root_dir
    old_pgsql_data_dir = old_pgsql_data_dir.replace('\\', '/')
    if os.path.isdir(old_pgsql_data_dir) and os.path.isdir(old_pgsql_root_dir):
        logger.info('Migrating database from previous 9.6 postgresql DB')
        migrate_pg_db(old_pgsql_root_dir, old_pgsql_data_dir, pgsql_root_dir, pgsql_data_dir)

    configure_max_connections(pgsql_data_dir)

    logger.info('Starting postgresql')
    if not setuphelpers.service_is_running('waptpostgresql'):
        setuphelpers.service_start('waptpostgresql')
        # waiting for postgres to be ready
        time.sleep(2)

    logger.info("Checking '%(db_name)s' database" % conf)
    import psycopg2
    from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
    conn = None
    cur = None
    try:
        conn = psycopg2.connect('dbname=template1 user=postgres')
        conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        cur = conn.cursor()
        cur.execute("select 1 from pg_roles where rolname='%(db_user)s'" % conf)
        val = cur.fetchone()
        if val is None:
            logger.info("%(db_user)s pgsql user does not exists, creating %(db_user)s user" % conf)
            cur.execute("create user %(db_user)s" % conf)

        cur.execute("select 1 from pg_database where datname='%(db_name)s'" % conf)
        val = cur.fetchone()
        if val is None:
            logger.info("Database %(db_name)s does not exists, creating %(db_name)s db" % conf)
            cur.execute("create database %(db_name)s owner %(db_user)s" % conf)

    finally:
        if cur:
            cur.close()
        if conn:
            conn.close()

    logger.info("Creating/upgrading wapt db tables")
    run(r'"%s\wapt-get.exe" "%s\waptserver\model.py" init_db -c "%s"' % (wapt_root_dir, wapt_root_dir, options.configfile))
    logger.info("Done")

    logger.info('Import local Packages data into database')

    repo = WaptLocalRepo(conf['wapt_folder'])
    load_db_config(conf)
    Packages.update_from_repo(repo)


def install_waptserver_service(options, conf=None, server_names = None):
    if setuphelpers.service_installed('WAPTServer'):
        if setuphelpers.service_is_running('WAPTServer'):
            setuphelpers.service_stop('WAPTServer')
        setuphelpers.service_delete('WAPTServer')

    if conf is None:
        conf = waptserver.config.load_config(options.configfile)

    conf_dir = os.path.join(wapt_root_dir, 'conf')
    if not os.path.isdir(conf_dir):
        os.makedirs(conf_dir)

    if not conf.get('server_uuid'):
        conf['server_uuid'] = str(uuid.uuid4())
        waptserver.config.write_config_file(options.configfile, conf)

    logger.info("Install waptserver")
    #service_binary = os.path.abspath(os.path.join(wapt_root_dir, 'waptservice.exe'))
    # make a symlink for waptsvc.exe -> waptserver.exe
    service_link = os.path.abspath(os.path.join(wapt_root_dir, 'waptserver.exe'))
    #if os.path.exists(service_link):
    #    os.unlink(service_link)
    #os.symlink(service_binary,service_link)

    if setuphelpers.service_installed('WAPTServer'):
        if setuphelpers.service_is_running('WAPTServer'):
            setuphelpers.service_stop('WAPTServer')
        setuphelpers.service_delete('WAPTServer')

    # install service
    run('"%s" --install' % (service_link,))

    tasks_db = os.path.join(wapt_root_dir, 'db')
    mkdir_p(tasks_db)

    if not conf.get('secret_key'):
        conf['secret_key'] = ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(64))
        waptserver.config.write_config_file(options.configfile, conf)

    if not conf.get('token_secret_key'):
        conf['token_secret_key'] = ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(64))
        waptserver.config.write_config_file(options.configfile, conf)

    if options.setpassword:
        conf['wapt_password'] = pbkdf2_sha256.hash(base64.b64decode(options.setpassword))
        waptserver.config.write_config_file(options.configfile, conf)

    clients_signing_certificate = conf.get('clients_signing_certificate')
    clients_signing_key = conf.get('clients_signing_key')
    clients_signing_crl = conf.get('clients_signing_crl')

    if not clients_signing_certificate or not clients_signing_key:
        clients_signing_certificate = os.path.join(wapt_root_dir, 'conf', 'ca-%s.crt' % server_names[0])
        clients_signing_key = os.path.join(wapt_root_dir, 'conf', 'ca-%s.pem' % server_names[0])

        conf['clients_signing_certificate'] = clients_signing_certificate
        conf['clients_signing_key'] = clients_signing_key
        waptserver.config.write_config_file(options.configfile, conf)

    if clients_signing_certificate is not None and clients_signing_key is not None and not os.path.isfile(clients_signing_certificate):
        logger.info('Create a certificate and key for clients certificate signing')

        key = SSLPrivateKey(clients_signing_key)
        if not os.path.isfile(clients_signing_key):
            logger.info('Create SSL RSA Key %s' % clients_signing_key)
            key.create()
            key.save_as_pem()

        crt = key.build_sign_certificate(cn='ca-%s' % server_names[0], is_code_signing=False, is_ca=True)
        logger.info('Create X509 cert %s' % clients_signing_certificate)
        crt.save_as_pem(clients_signing_certificate)

    # known certificates
    ssl_dir = conf['known_certificates_folder']
    if not os.path.isdir(ssl_dir):
        # move existing ssl dir in wapt repo to parent dir (default location)
        if os.path.isdir(os.path.join(conf['wapt_folder'], 'ssl')):
            shutil.move(os.path.join(conf['wapt_folder'], 'ssl'), ssl_dir)
        else:
            os.makedirs(ssl_dir)

    if clients_signing_certificate is not None and clients_signing_key is not None and clients_signing_crl is not None and not os.path.isfile(clients_signing_crl):
        logger.info('Create a CRL for clients certificate signing')
        key = SSLPrivateKey(clients_signing_key)
        crt = SSLCertificate(clients_signing_certificate)
        crl = SSLCRL(clients_signing_crl, cacert=crt, cakey=key)
        crl.revoke_cert()    # ensure Packages index
        crl.save_as_pem()

    update_all_crls(conf)

    repo = WaptLocalRepo(conf['wapt_folder'])
    repo.update_packages_index()

    # Rename drivers conf for new version
    wads_folder = conf['wads_folder']
    if os.path.isdir(wads_folder):
        for driver_bundle_filename in glob.glob(r"%s\drivers\conf\*" % wads_folder):
            if not os.path.isfile(driver_bundle_filename):
                continue
            with open(driver_bundle_filename) as f:
                data = json.loads(f.read())
            hash_bundle_filename = hashlib.sha256(data['name'].encode('utf-8')).hexdigest()
            if driver_bundle_filename.rsplit('\\', 1)[1] != hash_bundle_filename:
                dstfilename = os.path.join(driver_bundle_filename.rsplit('\\', 1)[0], hash_bundle_filename)
                print('Move %s --> %s' % (driver_bundle_filename, dstfilename))
                shutil.move(driver_bundle_filename, dstfilename)

    run('"%s" "%s" /t /grant "*S-1-5-20":(OI)(CI)(M)' % (makepath(system32(), 'icacls.exe'), wapt_root_dir))


def install_wapttasks_service(options, conf=None):
    if setuphelpers.service_installed('WAPTTasks'):
        if setuphelpers.service_is_running('WAPTTasks'):
            setuphelpers.service_stop('WAPTTasks')
        setuphelpers.service_delete('WAPTTasks')

    if conf is None:
        conf = waptserver.config.load_config(options.configfile)
    logger.info("Install wapttasks")

    tasks_db = os.path.join(wapt_root_dir, 'db')
    setuphelpers.run(r'"%s" "%s" /grant  "*S-1-5-20":(OI)(CI)(M)' % (makepath(system32(), 'icacls.exe'), tasks_db))


if __name__ == '__main__':
    usage = """\
    %prog [-c configfile] [install_nginx install_postgresql install_waptserver] [--dhparam-key-size=SIZE]

    WAPT Server services setup.

    actions is either :
      <nothing> : run service in foreground
      install_nginx: recreate a nginx config file matching settings in waptserver.ini
      install_postgresql: create if needed the 'wapt' postgresql DB, and register the waptpostgresql service managed by waptsvc
      install_waptserver: register the waptserver and wapttasks services managed by waptsvc
      create-check-dhparam: create a dhparam.pem for nginx
      all: run 'install_nginx', 'install_postgresql', 'install_waptserver'

    """
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s')
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    setloglevel(logger,'info')
    log_directory = os.path.join(wapt_root_dir, 'log')
    if not os.path.exists(log_directory):
        os.mkdir(log_directory)
        run(r'"%s" "%s" /t /grant  "*S-1-5-20":(OI)(CI)(M)' % (makepath(system32(), 'icacls.exe'), os.path.join(wapt_root_dir, 'log')))

    config_filename = os.path.join(wapt_root_dir, 'conf', 'waptserver.ini')

    parser = OptionParser(usage=usage, version='winsetup.py ' + __version__)
    parser.add_option('--waptbasedir', dest='waptbasedir', default=wapt_root_dir,
                      help='Wapt installation base dir (default: %default)')
    parser.add_option('-c', '--config', dest='configfile', default=config_filename,
                      help='Config file full path (default: %default)')
    parser.add_option('-n', '--server-names', dest='server_names', default=None,
                      help='server name and ip for certificate CN and SubjectAltNames. Separator is a comma (default: %default)')

    parser.add_option('-l', '--loglevel', dest='loglevel', default=None, type='choice',
                      choices=['debug',   'warning', 'info', 'error', 'critical'],
                      metavar='LOGLEVEL', help='Loglevel (default: info)')
    parser.add_option('-d', '--devel', dest='devel', default=False, action='store_true',
                      help='Enable debug mode (for development only)')
    parser.add_option('-f', '--force', dest='force', default=False, action='store_true',
                      help='Force rewrite nginx config')
    parser.add_option('-p', '--setpassword', dest='setpassword', default=None,
                      help='Set wapt server admin password. Value must be encoded in base64 (default: %default)')
    parser.add_option('--logto', dest='logtofilename', default=os.path.join(log_directory,'waptserver_install.log'),
                      help='Path for logs (default: %default)')
    parser.add_option(
        '--dhparam-key-size',
        dest='dhparam_key_size',
        default=1024,
        metavar='NUMBER',
        type='int',
        help='Size for dhparam key')

    (options, args) = parser.parse_args()

    wapt_root_dir = options.waptbasedir

    if options.server_names is not None:
        server_names = [s.strip() for s in options.server_names.split(',')]
    else:
        server_names = None

    if options.loglevel is not None:
        setloglevel(logger, options.loglevel)

    conf = waptserver.config.load_config(options.configfile)

    if conf['wapt_folder'].endswith('\\') or conf['wapt_folder'].endswith('/'):
        conf['wapt_folder'] = conf['wapt_folder'][:-1]

    # delete previous log file to be able to check easily from innosetup if a CRITICAL error has been raised.
    if options.logtofilename and os.path.exists(options.logtofilename):
        os.remove(options.logtofilename)

    h = logging.FileHandler(filename=options.logtofilename)
    h.setFormatter(formatter)
    logger.addHandler(h)

    dhparam_key_size = options.dhparam_key_size

    if args == ['all']:
        args = ['install_postgresql', 'install_waptserver', 'install_nginx']

    try:
        for action in args:
            if action == 'install_postgresql':
                logger.info('Installing PostgreSQL as a service managed by itself')
                install_postgresql_service(options, conf)
            elif action == 'install_waptserver':
                logger.info('Installing WAPT Server as a service managed by waptsvc')
                install_waptserver_service(options, conf, server_names=server_names)
                install_wapttasks_service(options, conf)

                for j in ['sync.json','rules.json','licences.json']:
                    apath = os.path.join(conf['wapt_folder'], j)
                    if not os.path.isfile(apath):
                        with open(apath, 'w') as f:
                            if j in ('licences.json','rules.json'):
                                f.write('[]')
                            elif j in ('sync.json'):
                                f.write('{}')

            elif action == 'install_nginx':
                logger.info('Installing nginx as a service managed by waptsvc')
                install_nginx_service(options, conf, server_names=server_names)
            elif action == 'create-check-dhparam':
                logger.info('Create or check dhparam file')
                ap_ssl_dir = os.path.join(wapt_root_dir, 'waptserver', 'nginx', 'ssl')
                ap_dhparam = os.path.join(ap_ssl_dir, 'dhparam.pem')
                if not (os.path.isdir(ap_ssl_dir)):
                    os.makedirs(ap_ssl_dir)
                if not(os.path.isfile(ap_dhparam)):
                    with open(ap_dhparam, 'wb') as f:
                        f.write(create_dhparam(dhparam_key_size))

            # we save at each step to get updated calculated params.
            conf = waptserver.config.load_config(cfgfile=options.configfile)

        if setuphelpers.service_installed('WAPTServer'):
            if not setuphelpers.service_is_running('WAPTServer'):
                setuphelpers.service_start('WAPTServer')

    except Exception as e:
        logger.critical(traceback.format_exc())
        raise e
